{-# OPTIONS --allow-unsolved-metas #-}
module 1-IndFamilies where

  {- Courtesy of Conor McBride -}

data Bool : Set where
  false : Bool
  true  : Bool 

data ℕ : Set where
  zero : ℕ
  suc : ℕ → ℕ

{-# BUILTIN NATURAL ℕ #-}
  -- The derivative above allows us to use built-in
  -- abbreviations, e.g. 0 for zero, 1 for suc zero,
  -- 2 for suc (suc zero).

data List (A : Set) : Set where
  [] : List A
  _∷_ : A → List A → List A

infixr 5 _∷_

  -- any connection between ℕ and List A?

length : ∀ {A} → List A → ℕ
length = {!   !}

data _×_ (A B : Set) : Set where
  _,_ : A → B → A × B

infixr 2 _×_
infixr 4 _,_

-- Vectors: lists indexed by their lengths

data Vec (A : Set) : ℕ → Set where
 [] : Vec A zero
 _∷_ : {n : ℕ} → A → Vec A n → Vec A (suc n)

 -- While List defines a datatype inductively,
 -- Vec inductively defines a *family* of types
 --   Vec A 0, Vec A 1, Vec A 2 ....

ex1 : Vec ℕ 3
ex1 = 1 ∷ (2 ∷ (3 ∷ []))

head : ∀ {A}{n} → Vec A (suc n) → A
head xs = {!   !}

tail : ∀ {A}{n} → Vec A (suc n) → Vec A n
tail xs = {!   !}

zip : ∀{A B n} → Vec A n → Vec B n → Vec (A × B) n
zip xs ys = {!   !}

zip3 : ∀{A B C n} →
         Vec A n → Vec B n → Vec C n →
           Vec (A × B × C) n
zip3 = {!   !}

map : ∀{A B n} → (A → B) → Vec A n → Vec B n
map = {!   !}

-- concatenating vectors

_+_ : ℕ → ℕ → ℕ
zero + n = n
(suc m) + n = suc (m + n)

  -- The type of _++_ (append) is
  --  _++_ : {A : Set} {m : ℕ} {n : ℕ} → Vec A m → Vec A n → Vec A (m + n)
  -- but A must be a Set and m and n must be ℕ, since they are
  -- used by Vec.

  -- When the types can be inferred, we can use the "∀" notation:

_++_ : ∀ {A m n} → Vec A m → Vec A n → Vec A (m + n)
xs ++ ys = {!   !}

  -- What is it about _+_ which makes _++_ typecheck?

data Even : ℕ → Set where
  ze : Even zero
  sse : ∀ {n} → Even n → Even (suc (suc n))

6-even : Even 6
6-even = {!   !}

sse→e : ∀ {n} → Even (suc (suc n)) → Even n
sse→e = {!   !}

e+e→e : ∀ {m n} → Even m → Even n → Even (m + n)
e+e→e = {!   !}

data ⊥ : Set where

¬ : Set → Set
¬ P = P → ⊥

¬e2+n→¬en : ∀ {n} → ¬ (Even (suc (suc n))) → ¬ (Even n)
¬e2+n→¬en {n} ¬e2+n en = {!   !}

⊥-elim : ∀ {P : Set} → ⊥ → P
⊥-elim = {!   !}

data Odd : ℕ → Set where
  zo  : Odd (suc zero)
  sso : ∀ {n} → Odd n → Odd (suc (suc n))

¬even→odd : ∀ n → ¬ (Even n) → Odd n
¬even→odd n = {!   !}

---------------------------------------
-- Products and existential quantifier
--   We have to talk about them because we will use them later..

-- As we know, if (x , y) has type (A × B), x has type A
-- while y has type B.
-- In dependently typed language one can generalize this a bit:
-- the type of y may depend on the value of x.

data Σ (A : Set) (P : A → Set) : Set where
  _,_ : (x : A) → (y : P x) → Σ A P

-- A × B is a special case:
--  _×_ : (A : Set) (B : Set) → Set
--  A × B = Σ A (λ _ → B)

-- E.g. A function returning a type

BoolOrℕ : Bool → Set
BoolOrℕ false = Bool
BoolOrℕ true  = ℕ

bn0 : Σ Bool BoolOrℕ
bn0 = (false , ?)

bn1 : Σ Bool BoolOrℕ
bn1 = (true , ?)

-- The ∃ quantifier can be understood as a pair.

∃ : {A : Set} → (A → Set) → Set
∃ {A} P = Σ A P

-- That is, ∃ P is merely Σ A P with the A omitted for brevity.
-- Note that P is a predicate, that is, a function of type A → Set.

-- A proof of ∃ {A} P is a pair (x, pf)
--    x has type A, while pf is a proof of P x.

-- Let's prove something silly: there exists a number n
--  such that suc n is even.
-- In usual math notation you would write  
--  (∃ n ∈ ℕ : Even (suc n))

existsEven : ∃ (λ n → Even (suc n))
existsEven = {!   !}
